/*
 * $Id: rb_event_handler.c 358 2006-02-12 15:42:28Z tilman $
 *
 * Copyright (C) 2004 ruby-ecore team (see AUTHORS)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ruby.h>
#include <stdbool.h>

#include <Ecore.h>

#define __RB_EVENT_HANDLER_C
#include "rb_ecore.h"
#include "rb_event_handler.h"

typedef struct {
	Ecore_Event_Handler *real;
	VALUE callback;
	bool deleted;
} RbEventHandler;

static int on_ecore_event (void *data, int type, void *event);

VALUE event_classes, cEcoreEvent;
static VALUE handlers;

static void c_mark (RbEventHandler *h)
{
	rb_gc_mark (h->callback);
}

static void c_free (RbEventHandler *h)
{
	RbEventHandler *h2 = NULL;
	int len = RARRAY (handlers)->len, i;
	VALUE el;

	if (!h->real || h->deleted) {
		free (h);
		return;
	}

	for (i = 0; i < len; i++) {
		el = rb_ary_shift (handlers);

		Data_Get_Struct (el, RbEventHandler, h2);

		if (h == h2) {
			ecore_event_handler_del (h->real);
			break;
		}

		rb_ary_push (handlers, el);
	}

	free (h);
}

static VALUE c_alloc (VALUE klass)
{
	RbEventHandler *h = NULL;

	return Data_Make_Struct (klass, RbEventHandler, c_mark, c_free, h);
}

static VALUE c_init (VALUE self, VALUE type)
{
	RbEventHandler *h = NULL;
	int t;

	if (!rb_block_given_p ())
		rb_raise (rb_eStandardError, "block missing");

	Data_Get_Struct (self, RbEventHandler, h);

	t = NUM2INT (type);

	if (t <= ECORE_EVENT_NONE)
		rb_raise (rb_eStandardError, "invalid type");

	rb_iv_set (self, "@type", type);

	h->callback = rb_block_proc ();
	h->deleted = false;
	h->real = ecore_event_handler_add (t, on_ecore_event, NULL);

	rb_ary_push (handlers, self);

	return self;
}

static VALUE c_delete (VALUE self)
{
	RbEventHandler *h = NULL;
	int len = RARRAY (handlers)->len, i;
	VALUE el;

	for (i = 0; i < len; i++) {
		el = rb_ary_shift (handlers);
		if (el == self) {
			Data_Get_Struct (self, RbEventHandler, h);
			ecore_event_handler_del (h->real);
			h->real = NULL;
			h->deleted = true;

			return Qnil;
		}

		rb_ary_push (handlers, el);
	}

	return Qnil;
}

static int on_ecore_event (void *data, int type, void *event)
{
	RbEventHandler *h = NULL;
	VALUE handler, klass, obj, argv[1], res;
	int handler_type, len, i;
	bool called = false;

	/* instantiate the event object
	 * first, find the class we're gonna use
	 */
	if (NIL_P (klass = rb_hash_aref (event_classes, INT2NUM (type)))) {
		rb_raise (rb_eException, "Cannot find event class "
		                         "for event %i\n", type);
		return 0;
	}

	/* now create and init the object */
	obj = rb_obj_alloc (klass);
	argv[0] = (VALUE) event;
	rb_obj_call_init (obj, 1, argv);

	len = RARRAY (handlers)->len;

	for (i = 0; i < len; i++) {
		handler = rb_ary_entry (handlers, i);
		handler_type = NUM2INT (rb_iv_get (handler, "@type"));

		if (handler_type == type) {
			Data_Get_Struct (handler, RbEventHandler, h);
			res = rb_funcall (h->callback, rb_intern ("call"), 1, obj);
			called = true;

			/* if the block returned false, don't call the other
			 * event handlers
			 */
			if (res == Qfalse)
				break;
		}
	}

	if (type == ECORE_EVENT_SIGNAL_EXIT && !called)
		ecore_main_loop_quit ();

	/* call other event handlers, too */
	return 1;
}

VALUE c_ev_generic_init (VALUE self, VALUE event)
{
	/* dummy */
	return self;
}

void Init_EventHandler (void)
{
	VALUE cEventHandler;

	cEventHandler = rb_define_class_under (mEcore, "EventHandler",
	                                       rb_cObject);

	rb_define_alloc_func (cEventHandler, c_alloc);
	rb_define_method (cEventHandler, "initialize", c_init, 1);
	rb_define_method (cEventHandler, "delete", c_delete, 0);

	handlers = rb_ary_new ();
	rb_global_variable (&handlers);

	event_classes = rb_hash_new ();
	rb_global_variable (&event_classes);

	/* define a base event class */
	cEcoreEvent = rb_define_class_under (mEcore, "Event", rb_cObject);
	rb_define_private_method (rb_singleton_class (cEcoreEvent),
	                          "new", NULL, 0);
}
